#ifndef MSP_GEOMETRY_HYPERSPHERE_H_
#define MSP_GEOMETRY_HYPERSPHERE_H_

#include <cmath>
#include <stdexcept>
#include <msp/linal/vector.h>
#include "shape.h"

namespace Msp {
namespace Geometry {

/**
A shape consisting of the points within a specific distance from the origin.
Two- and three-dimensional cases are Circle and Sphere, respectively.
*/
template<typename T, unsigned D>
class HyperSphere: public Shape<T, D>
{
private:
	T radius;

public:
	HyperSphere(): radius(1) { }
	explicit HyperSphere(T);

	virtual HyperSphere *clone() const;

	T get_radius() const { return radius; }

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box() const;
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return 2; }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
};

template<typename T, unsigned D>
inline HyperSphere<T, D>::HyperSphere(T r):
	radius(r)
{
	if(r<=T(0))
		throw std::invalid_argument("HyperSphere::HyperShpere");
}

template<typename T, unsigned D>
inline HyperSphere<T, D> *HyperSphere<T, D>::clone() const
{
	return new HyperSphere<T, D>(radius);
}

template<typename T, unsigned D>
inline BoundingBox<T, D> HyperSphere<T, D>::get_axis_aligned_bounding_box() const
{
	LinAl::Vector<T, D> extent;
	for(unsigned i=0; i<D; ++i)
		extent[i] = radius;
	return BoundingBox<T, D>(-extent, extent);
}

template<typename T, unsigned D>
inline bool HyperSphere<T, D>::contains(const LinAl::Vector<T, D> &point) const
{
	return inner_product(point, point)<=radius*radius;
}

template<typename T, unsigned D>
inline unsigned HyperSphere<T, D>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	using std::sqrt;

	T mid = -inner_product(ray.get_direction(), ray.get_start());
	LinAl::Vector<T, D> nearest = ray.get_start()+ray.get_direction()*mid;
	T offset_sq = radius*radius-inner_product(nearest, nearest);
	if(offset_sq<0)
		return 0;
	T offset = sqrt(offset_sq);

	unsigned n = 0;
	for(int i=-1; (n<size && i<=1); i+=2)
	{
		T x = mid+offset*i;
		if(ray.check_limits(x))
		{
			if(points)
			{
				points[n].position = ray.get_start()+ray.get_direction()*x;
				points[n].normal = normalize(points[n].position);
				points[n].distance = x;
			}

			++n;
		}
	}

	return n;
}

} // namespace Geometry
} // namespace Msp

#endif
